"""
    Code for measuring OPC-UA PubSub elapsed time over 5G using Cloudified PLC
    Closed Control Loop Latency data is exported to a CSV file
"""

from datetime import datetime
from asyncua import Client, ua, Node
import asyncio
import csv
import random

#OPCUA Namespace
namespace = "OPC-UA_Server"
#Period PUBSUB (ms) {1,10,50}
period_ms = 0
#CSV Filename
filename = ""

#Number of bytes
n_bytes = 0
#Txt if the packet is 1500 bytes
text1500 = ""
#Value read
read_value = ""
#Number of packets
n_packets = 0

#Dict for before packet times
before_times = {}
for i in range(0,n_packets):
    before_times[i] = datetime.now()

#Dict for after packet times
after_times = {}
for i in range(0,n_packets):
    after_times[i] = datetime.now()

#Dict for elapsed  times
diff_times = {}
for i in range(0,n_packets):
    diff_times[i] = -1.0



#logging.basicConfig(level=logging.WARNING)

class SubHandler:

    #Function initialization (constructor)
    def __init__(self, aux_value: Node):
        self.aux_value = aux_value


    #Function that tells us about the change in data
    async def datachange_notification(self, node, val, attr):
        global after_times, read_value
       
        print("NODE 2 HAS MADE A CHANGE IN RESPONSE")
        #We get the timestamp
        end_time = datetime.now()
        #If the size of the payload is aprox 1500, we split the huge msg and get the number of the packet
        #If not, the number of the packet is the read value
        if n_bytes == 1300: 
            aux = val.split(";")
            read_value = aux[1]
        else:
            read_value = val        
        #We store the timestamp in the dict
        after_times[int(read_value)] = end_time
    
        #print("New data change event:")
        #print("Node: ", node)
        #print("Value: ", read_value)


#We separate WRITE of SUBS in 2 functions because that way the async waits for the last to complete
# so its a concurrent/parallel scenario
# |----------------------- WRITE ENDS                         | From here both finished but we wait for the last to complete
# |--------------------------------------------- SUBS ENDS    |

#Function for writing a value (val) to the node (variable) in the server
async def write_message(variable:Node):
    global before_times, n_bytes

    #We study the data type so we can be aware at the time of writing the message
    for i in range(0,n_packets):
        #We get the data type of the variable from the server
        data_type = await variable.read_data_type_as_variant_type()
        if str(data_type) == "VariantType.String":
            var_type = ua.VariantType.String
            #If the packet size is 2 bytes we send the number of the packet
            if n_bytes == 2:
                payload = str(i)
            #Id the packet size is around 1300 bytes we send the full message + ; + number of packet
            elif n_bytes == 1300:
                payload = text1500+";"+str(i)

        elif str(data_type) == "VariantType.Int64":
            #print("WE ARE IN 2 BYTES INT")
            payload = i
            var_type = ua.VariantType.Int64
        
        #Get the timestamp before writting the value
        write_time = datetime.now()
        #We store the timestamp in the dict
        before_times[i] = write_time
        #We write in the node
        await variable.write_value(payload,var_type)
        print("NODE 1 ---> WRITES VALUE ON REQUEST")
        #We sleep about 15-40 ms so we can wait after every packet has been sent
        await asyncio.sleep(random.uniform(0.015,0.04))


#Function for the SUBSCRIPTION
async def subscription(client: Client, variable_sub:Node, variable_pub:Node):
    #We create a SUBSCRIPTION 
    sub = await client.create_subscription(period_ms, SubHandler(variable_sub))
    #We subscribe to the value change of the variable temperature
    handle_var = await sub.subscribe_data_change(variable_sub)
    print("NODE 1 ---> SUBSCRIBED TO RESPONSE")
    #We write the value
    await write_message(variable_pub)
    #We wait all packets are done before unsubscribing
    await asyncio.sleep(3)
    await sub.unsubscribe(handle_var)
    await sub.delete()


async def main():
    global diff_times, n_packets, period_ms, filename, n_bytes, text1500

    #We ask for the IP of the server
    IP_server = input("OPC-UA's Server IP: ")
    #We connect to the OPC-UA server
    client = Client("opc.tcp://"+IP_server+":4840/freeopcua/server/")
    try:
        await client.connect()

        #We specify the payload size
        payload_size = input("Payload size: \na) 2 bytes\nb) aprox 1300 bytes\n: ")
        payload_size = payload_size.lower()
        #If the packet size is 1500 bytes, we generate that payload
        if payload_size == "b":
            n_bytes = 1300
            with open('1500bytes.txt', 'r') as file:
                # Read all the lines of the file into a list
                lines = file.readlines()
                # Join the lines into a single string
                text1500 = ''.join(lines)
        else:
            n_bytes = 2
            
        #We specify the loaded network
        bw_used = int(input("Load of the network: "))
        #We specify the number of packets
        n_packets = int(input("Number of packets: "))
        #We specify the pubsub period
        period_ms = int(input("PubSub period {1,10,50}: "))
        #We get the namespace of the node
        namespaceidx = await client.get_namespace_index(namespace)
        #variable we want to subscribe
        variable_sub =  await client.nodes.root.get_child(
                        ["0:Objects", f"{namespaceidx}:PLC", f"{namespaceidx}:Response"])
        #Variable we want to change
        variable_pub =  await client.nodes.root.get_child(
                        ["0:Objects", f"{namespaceidx}:PLC", f"{namespaceidx}:Request"])
        
        #We SUBSCRIBE
        await subscription(client,variable_sub,variable_pub)

        #We specify the filename
        filename = "CCLL_new_5G_TESTBED_"+str(n_bytes)+"bytes_"+str(period_ms)+"ms_load"+str(bw_used)+".csv"
        
        #We open the CSV file and write all the data on it
        with open(filename, "w+", newline='') as file:
            writer = csv.writer(file)
            #We write the header
            writer.writerow(["TX Time", "RX Time", "Diff Time"])
            #Once all the process SUBSCRIBE-WRITE is finished, we compute the elapsed times
            for i in range(0,n_packets):
                diff_times[i] = after_times[i] - before_times[i]
                #We get the seconds and microsec of the closed loop delay
                seconds = diff_times[i].total_seconds()
                milliseconds = seconds*1000
                print("TIME ELAPSED FOR PACKET ", read_value,": ", milliseconds)
                #We store the data
                row_data = [before_times[i], after_times[i], milliseconds]
                #We store it in the CSV
                writer.writerow(row_data)
           
    finally:
        await client.disconnect()

if __name__ == "__main__":
    asyncio.run(main())
