"""
Code used for moving the robotic arm from the Edge Cloud (Cloudified PLC)
Output: a CSV is created storing the total cycle elapsed rime of the whole routine
"""

import logging
from asyncua import Client, Node
import asyncio
import time
import csv

#OPCUA Namespace
namespace = "OPC-UA_Server"
#Period PUBSUB (ms)
period_ms = 10

current_position = ""

#Number of times to do the routine
n_times = 5
#Filename
filename = ""
#List for storing the elapsed times
time_list = []

logging.basicConfig(level=logging.WARNING)

class SubHandler:

    #Function initialization (constructor)
    def __init__(self, aux_value: Node):
        self.aux_value = aux_value


    #Function that tells us about the change in data
    async def datachange_notification(self, node, val, attr):
        global current_position
        #We see what is the new data value (it changed)
        """
        print("New data change event:")
        print("Node: ", node)
        print("Value: ", val)
        """
        #We store oour current position
        current_position = val

#Function for the SUBSCRIPTION
async def subscription(client: Client, variable:Node):

    #We create a SUBSCRIPTION.
    sub = await client.create_subscription(period_ms, SubHandler(variable))
    #We subscribe to the value change of the variable temperature
    handle_var = await sub.subscribe_data_change(variable)

    return sub, handle_var


#Function to move the robotic arm
#It can be changed in order to try different movements
async def move_1(node):
    global time_list
    #Array of all positions
    all_positions = []
    all_positions.append([0.125664, -2.07083, -2.20086, -0.39235, 1.459793, 0.890642])
    all_positions.append([0.125664, -1.80903, -2.20086, -0.39235, 1.459793, 0.890642])
    all_positions.append([-1.51058, -1.80903, -2.20086, -0.39235, 1.459793, 0.890642])
    all_positions.append([-1.51058, -1.80903, -1.102, -0.39235, 1.459793, 0.890642])
    all_positions.append([-1.51058, -1.80903, -1.102, -0.23754, 1.459793, 0.890642])

    start_time = time.time()
    #We loop the positions ascending order
    for i in range(0,len(all_positions)):
        await node.write_value("[{}, {}, {}, {}, {}, {}]".format(all_positions[i][0],all_positions[i][1],all_positions[i][2],all_positions[i][3],all_positions[i][4],all_positions[i][5]))
        await wait_pos(all_positions[i])
                    
    #We loop the positions descending (to reverse the movement)
    for i in range(0,len(all_positions)):
        await node.write_value("[{}, {}, {}, {}, {}, {}]".format(all_positions[len(all_positions)-i-1][0],all_positions[len(all_positions)-i-1][1],all_positions[len(all_positions)-i-1][2],all_positions[len(all_positions)-i-1][3],all_positions[len(all_positions)-i-1][4],all_positions[len(all_positions)-i-1][5]))
        await wait_pos(all_positions[len(all_positions)-i-1])

    end_time = time.time()
    elapsed_time = end_time-start_time
    print("ELAPSED TIME OPC-UA: ", round(elapsed_time,3), " SECONDS")
    time_list.append(elapsed_time)



#Function used to convert a string list to normal list (form "[12, 45, 23]" to [12,45,23])
async def convert_str_list(data_input: str):
    data_output = data_input.replace(" ", "")
    data_output = data_output.replace("[", "")
    data_output = data_output.replace("]", "")
    output = data_output.split(",")

    converted_list = []
    for i in range(0, len(output)):
        converted_list.append(float(output[i]))
    return converted_list


#Function to waint until the position is reached
async def wait_pos(dest_pos):

    #Our threshold is 0.01*pi/180 = 0.000175 (this is our error of position in radians of every joint)
    #This is because the robot joint usually changes the 0.01 degree, it is not stable 
    #We will do it with the threshold 5 times
    threshold = 0.000175*5
    
    while True:
        await asyncio.sleep(0.001)
        #We get the current OPCUA position of the joints and convert it to list
        #print("CURRENT POS B4 CONVERT: ", current_position)
        current_opcua_position = await convert_str_list(current_position)
        #print("CURRENT POS: ", current_opcua_position)
        #print("DESTINATION POS: ", dest_pos)
        #Flag to exit to the while True level
        flag = True

        for i in range(0,len(dest_pos)):
            result = abs(dest_pos[i]-current_opcua_position[i])
            if result > threshold:
                flag = False
                break
        
        if flag == True:
            #print("reach")
            break
        #print("not reach")
                
            

#Function used to change the of the variable joints
async def change_value(node: Node):
    
    #while True:
    for i in range(0,n_times):   
        print("========================")
        print("Iteration ",i)
        await move_1(node)
        print("========================")

  

async def main():
    global n_times, filename
    #We ask which arm to connect
    #arm_IP = input("Enter the IP of the arm: ")
    arm_IP = "192.168.100.32"
    #We ask for the times
    #n_times = int(input("Number of times: "))
    #Filename
    filename = "TIME_ELAPSED_DEMO_OPCUA_"+str(n_times)+"_TIMES.csv"
    #We connect to the OPC-UA server
    client = Client("opc.tcp://"+arm_IP+":4840/freeopcua/server/")
    try:
        await client.connect()
    
        #We get the namespace of the node
        namespaceidx = await client.get_namespace_index(namespace)
        #We get the childs (nodes) of the namespace
        var_joints =  await client.nodes.root.get_child(
                        ["0:Objects", f"{namespaceidx}:UR_Arm", f"{namespaceidx}:Joint_Values"])

        #We get the current values of the joints 
        var_current =  await client.nodes.root.get_child(
                        ["0:Objects", f"{namespaceidx}:UR_Arm", f"{namespaceidx}:Current_Joints"])
      
      
        #We subscribe to the current values of the joints -> variable in OPCUA server called Current_Joints
        #We change value / write the joints using the OPCUA variable called Joint_Values
        sub, handle_var = await subscription(client,var_current)
        await change_value(var_joints)

        #We write data in the CSV
        with open(filename, 'w', newline='') as csvfile:
            writer = csv.writer(csvfile)
            # Write the headers
            writer.writerow(['Iteration', 'ElapsedTime'])
            # Write the list elements
            for index, elapsed_time in enumerate(time_list, start=1):
                writer.writerow([index, elapsed_time])
        
        print("CSV file created and written!")
        print("Test has ended!")

      
    finally:
        await sub.unsubscribe(handle_var)
        await sub.delete()
        await client.disconnect()
        

if __name__ == "__main__":
    asyncio.run(main())
