"""
Code used for moving the robotic arm from the Edge Cloud (Cloudified PLC)
Robotic cell side
"""

import rtde.rtde as rtde
import rtde.rtde_config as rtde_config

import asyncio
from asyncua import Server,ua, Node
import logging
import sys
import ast
import json
from collections import namedtuple
from threading import Thread


#Refreshing period of PubSub
period_ms = 10
#Current position of the arm
current_RTDE = None


class SubHandler:

    #Function initialization (constructor)
    def __init__(self, aux_value: Node):
        self.aux_value = aux_value


    #Function that tells us about the change in data
    async def datachange_notification(self, node, val, attr):
        #añadido 13/6/2023
        global motion
        #We see what is the new data value (it changed)
        print("New data change event:")
        print("Node: ", node)
        print("Value: ", val)

        #Add here a before dict and the key = motion (data RX from edge)
        motion = ast.literal_eval(val)
        print("VALUE OF JOINTS: ", motion)
        await write_RTDE(motion)



#Function for the SUBSCRIPTION
async def subscription(server: Server, variable:Node):

    #We create a SUBSCRIPTION.
    sub = await server.create_subscription(period_ms, SubHandler(variable))
    #We subscribe to the value change of the variable temperature
    handle_var = await sub.subscribe_data_change(variable)
    #Once we are subscribed, we change the value of the variable
    return sub, handle_var



#We start the motion variable to none 
# before getting data from RTDE we need to have something even though it is empty
#Initial position
motion = [0,0,0,0,0,0]


# logging.basicConfig(level=logging.INFO)
ROBOT_HOST = "127.0.0.1"
ROBOT_PORT = 30004
config_filename = "control_loop_configuration.xml"


logging.getLogger().setLevel(logging.INFO)

conf = rtde_config.ConfigFile(config_filename)
state_names, state_types = conf.get_recipe("state")
actualq_names, actualq_types = conf.get_recipe("actualq")
#watchdog_names, watchdog_types = conf.get_recipe("watchdog")

con = rtde.RTDE(ROBOT_HOST, ROBOT_PORT)
con.connect()
print("Connected to RTDE")

# get controller version
con.get_controller_version()

# setup recipes
con.send_output_setup(state_names, state_types)
actualq = con.send_input_setup(actualq_names, actualq_types)


#We apply the initial position to the joints
actualq.input_double_register_24 = motion[0]
actualq.input_double_register_25 = motion[1]
actualq.input_double_register_26 = motion[2]
actualq.input_double_register_27 = motion[3]
actualq.input_double_register_28 = motion[4]
actualq.input_double_register_29 = motion[5] 


#Function used to read the data of the joints using RTDE
def read_RTDE():
    global current_RTDE
    conf = rtde_config.ConfigFile("record_configuration.xml")
    output_names, output_types = conf.get_recipe("out")

    # get controller version
    con.get_controller_version()

    # setup recipes
    if not con.send_output_setup(output_names, output_types, frequency=50):
        logging.error("Unable to configure output")
        sys.exit()

    # start data synchronization
    if not con.send_start():
        logging.error("Unable to start synchronization")
        sys.exit()
   

    keep_running = True
    while keep_running:
        try:
            
            #Getting data from the joints from RTDE
            state = con.receive_buffered()
            if state is not None:
                #We convert the data into json object
                data = json.dumps(state.__dict__) #serialize data object
                #Converting JSON data into Python readable object
                current_RTDE = json.loads(data, object_hook = lambda d : namedtuple('a', d.keys()) (*d.values()))

        except KeyboardInterrupt:
            keep_running = False
        except rtde.RTDEException:
            con.disconnect()
            sys.exit()

    con.send_pause()
    con.disconnect()


#Function used to write the joint with the values we get from the edge cloud
async def write_RTDE(motion):

    actualq.input_double_register_24 = motion[0]
    actualq.input_double_register_25 = motion[1]
    actualq.input_double_register_26 = motion[2]
    actualq.input_double_register_27 = motion[3]
    actualq.input_double_register_28 = motion[4]
    actualq.input_double_register_29 = motion[5] 
        
    con.send(actualq)
    print("Se ha enviado a RTDE el comando")
    print(motion)
    

async def main():
    _logger = logging.getLogger(__name__)
    # set up our server
    server = Server()
    await server.init()
    server.set_endpoint("opc.tcp://0.0.0.0:4840/freeopcua/server/")

    # set up our own namespace, not really necessary but should as spec
    # uri = "http://examples.freeopcua.github.io"
    uri = "OPC-UA_Server"
    idx = await server.register_namespace(uri)
    #Log messages
    _logger.info("idx is {}".format(idx))

    # server.nodes, contains, links to very common nodes like objects and root
    #We creatre a Node inside the namespace of this server
    arm_node = await server.nodes.objects.add_object(idx, "UR_Arm")
    #Log message
    _logger.info("myobj is {}".format(arm_node))
    #We add one variable to the node
    joint_values = await arm_node.add_variable(idx, "Joint_Values", "[0,0,0,0,0,0]",ua.VariantType.String)
    #Log message
    _logger.info("myvar is {}".format(joint_values))
    #Variable for knowing current position
    current_joints = await arm_node.add_variable(idx, "Current_Joints", "[0,0,0,0,0,0]",ua.VariantType.String)
    _logger.info("myvar is {}".format(current_joints))

    # Set both variables to be writable
    await joint_values.set_writable()
    await current_joints.set_writable()

    #We subscribe to the joint_values (the ones the edge cloud client is writing to)
    sub_joint, sub_handle_var = await subscription(server,joint_values)
    _logger.info("Starting server!")

    async with server:
        try:
            #We create an old list to know the old position
            old_current_pos_list = ""
            while True:
               
                await asyncio.sleep(0.005)
                #We store the current position
                current_pos_list = "["+str(current_RTDE.actual_q[0])+","+str(current_RTDE.actual_q[1])+"," + str(current_RTDE.actual_q[2])+"," + str(current_RTDE.actual_q[3]) + "," + str(current_RTDE.actual_q[4]) + "," + str(current_RTDE.actual_q[5])+"]"
                #If our last position (old) and the current one are different, we write the last position in OPC-UA server
                if old_current_pos_list != current_pos_list:
                    #Here add the after timestamp
                    """
                    here we compare the current_pos_list with motion (every joint) comparing with a threshold
                    IF ALL OF THEM FIT, THEN:
                    after_times[global key] = datetime.now()

                    Use an i or idx as a global variable because I know it is from 0 to 509 (510 samples)
                    After saving the after_times, we do i+=1 (endo of one loop)
                    """
                    current_values = await current_joints.write_value(current_pos_list,ua.VariantType.String)
                    old_current_pos_list = current_pos_list

        #We finish the server if there is an exception (keyboard)        
        except (KeyboardInterrupt, SystemExit):
            await sub_joint.unsubscribe(sub_handle_var)
            await sub_joint.delete()
            await server.stop()
            sys.exit()



if __name__ == "__main__":
    #If we set level = logging.DEBUG there is a lot of information
    #If we set level = logging.INFO we get less info
    logging.basicConfig(format="%(asctime)-15s %(message)s",
                        level=logging.INFO)
    

    #Thread for executing the RTDE server while reading the joint values
    thread_RTDE = Thread(target=read_RTDE)
    thread_RTDE.start()
    #Other thread for the main function (because we are always publishing in the server)
    asyncio.run(main(), debug=True)